@@ -32,7 +32,7 @@ use URI;
 use base 'Exporter';
 
 BEGIN {
-   $VERSION = 1.1;
+   $VERSION = 1.2;
    @EXPORT_OK = qw(
          format_source dumpval sv_peek sv_dump
          digest
@@ -108,6 +108,30 @@ sub dumpval {
    } || "[unable to dump $_[0]: '$@']";
 }
 
+=item $ref = from_json $json
+
+Converts a JSON string into the corresponding perl data structure.
+
+=cut
+
+sub from_json($) {
+   require JSON::Syck;
+   local $JSON::Syck::ImplicitUnicode = 1; # work around JSON::Syck bugs
+   JSON::Syck::Load ($_[0])
+}
+
+=item $json = to_json $ref
+
+Converts a perl data structure into its JSON representation.
+
+=cut
+
+sub to_json($) {
+   require JSON::Syck;
+   local $JSON::Syck::ImplicitUnicode = 0;  # work around JSON::Syck bugs
+   JSON::Syck::Dump ($_[0])
+}
+
 =item digest(args...)
 
 Calculate a SHA1 digest and return it base64-encoded. The result will
